/*-----------------------------------------------------------------------------
 * Name:    GLCD_MCB1800.c
 * Purpose: Graphic LCD interface (320x240 pixels) for Graphic LCD with
 *          SPI interface (in RGB mode) for EFM32GG-DK3750 Development Kit
 * Rev.:    1.0.0
 *----------------------------------------------------------------------------*/

/* Copyright (c) 2013 - 2018 ARM LIMITED

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
   ---------------------------------------------------------------------------*/

#include "RTE_Components.h"

#include "bsp.h"
#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_EBI
#include "tftamapped.h"
#endif
#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_RGB
#include "tftdirect.h"
#endif
#include "glib.h"

#include "GLCD_Config.h"
#include "Board_GLCD.h"

static GLIB_Context_t gc;
static uint16_t      *frameBuffer = NULL;
static GLCD_FONT     *pFont       = NULL;

#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_RGB
/* Configure TFT direct drive from EBI BANK2 */
const EBI_TFTInit_TypeDef tftInit =
{ ebiTFTBank2,                  /* Select EBI Bank 2 */
  ebiTFTWidthHalfWord,          /* Select 2-byte (16-bit RGB565) increments */
  ebiTFTColorSrcMem,            /* Use memory as source for mask/blending */
  ebiTFTInterleaveUnlimited,    /* Unlimited interleaved accesses */
  ebiTFTFrameBufTriggerVSync,   /* VSYNC as frame buffer update trigger */
  false,                        /* Drive DCLK from negative edge of internal clock */
  ebiTFTMBDisabled,             /* No masking and alpha blending enabled */
  ebiTFTDDModeExternal,         /* Drive from external memory */
  ebiActiveLow,                 /* CS Active Low polarity */
  ebiActiveHigh,                /* DCLK Active High polarity */
  ebiActiveLow,                 /* DATAEN Active Low polarity */
  ebiActiveLow,                 /* HSYNC Active Low polarity */
  ebiActiveLow,                 /* VSYNC Active Low polarity */
  320,                          /* Horizontal size in pixels */
  1,                            /* Horizontal Front Porch */
  30,                           /* Horizontal Back Porch */
  2,                            /* Horizontal Synchronization Pulse Width */
  240,                          /* Vertical size in pixels */
  1,                            /* Vertical Front Porch */
  4,                            /* Vertical Back Porch */
  2,                            /* Vertical Synchronization Pulse Width */
  0x0000,                       /* Frame Address pointer offset to EBI memory base */
  8,                            /* DCLK Period */
  0,                            /* DCLK Start cycles */
  0,                            /* DCLK Setup cycles */
  0,                            /* DCLK Hold cycles */
};
#endif


/**
  \fn          int32_t GLCD_Initialize (void)
  \brief       Initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Initialize (void) {
  EMSTATUS status;
  bool     init;

#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_EBI
  init = TFT_AddressMappedInit();
#endif
#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_RGB
  frameBuffer = (uint16_t *) EBI_BankAddress(EBI_BANK2);
  init = TFT_DirectInit(&tftInit);
#endif
  if (!init) {
    return -1;
  }

  status = GLIB_contextInit(&gc);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_Uninitialize (void)
  \brief       De-initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Uninitialize (void) {
  return -1;
}

/**
  \fn          int32_t GLCD_SetForegroundColor (uint32_t color)
  \brief       Set foreground color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetForegroundColor (uint32_t color) {
  gc.foregroundColor = color;
  return 0;
}

/**
  \fn          int32_t GLCD_SetBackgroundColor (uint32_t color)
  \brief       Set background color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetBackgroundColor (uint32_t color) {
  gc.backgroundColor = color;
  return 0;
}

/**
  \fn          int32_t GLCD_ClearScreen (void)
  \brief       Clear screen (with active background color)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_ClearScreen (void) {
  EMSTATUS status;

  status = GLIB_clear(&gc);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_SetFont (GLCD_FONT *font)
  \brief       Set active font
  \param[in]   font   Pointer to font structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetFont (GLCD_FONT *font) {
  pFont = font;
  return 0;
}

/**
  \fn          int32_t GLCD_DrawPixel (uint32_t x, uint32_t y)
  \brief       Draw pixel (in active foreground color)
  \param[in]   x  x position in pixels (0 = left corner)
  \param[in]   y  y position in pixels (0 = upper corner)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawPixel (uint32_t x, uint32_t y) {
  EMSTATUS status;

  status = GLIB_drawPixel(&gc, x, y);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw horizontal line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length) {
  EMSTATUS status;

  status = GLIB_drawLineH(&gc, x, y, x + length - 1);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw vertical line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length) {
  EMSTATUS status;

  status = GLIB_drawLineV(&gc, x, y, y + length - 1);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height)
  \brief       Draw rectangle (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Rectangle width in pixels
  \param[in]   height Rectangle height in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height) {
  GLIB_Rectangle_t rect; 
  EMSTATUS status;

  rect.xMin = x;
  rect.yMin = y;
  rect.xMax = x + width  - 1;
  rect.yMax = y + height - 1;
  status = GLIB_drawRect(&gc, &rect);
  if (status != GLIB_OK) {
    return -1;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch)
  \brief       Draw character (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   ch     Character
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch) {
  EMSTATUS status;
  uint32_t i, j;
  uint32_t wb, row;
  union {
    uint8_t  *p8;
    uint16_t *p16;
    uint32_t *p32;
  } bmp;

  if (pFont == NULL) return -1;

  ch -=  pFont->offset;
  wb  = (pFont->width + 7) / 8;

  switch (wb) {
    case 1:
      bmp.p8  = (uint8_t  *)(&pFont->bitmap[ch * wb * pFont->height]);
      break;
    case 2:
      bmp.p16 = (uint16_t *)(&pFont->bitmap[ch * wb * pFont->height]);
      break;
    case 4:
      bmp.p32 = (uint32_t *)(&pFont->bitmap[ch * wb * pFont->height]);
      break;
    default:
      return (-1);
  }

  for (j = 0; j < pFont->height; j++) {
    switch (wb) {
      case 1:
        row = *bmp.p8++;
        break;
      case 2:
        row = *bmp.p16++;
        break;
      case 4:
        row = *bmp.p32++;
        break;
    }
    for (i = 0; i < pFont->width; i++) {
      status = GLIB_drawPixelColor(&gc, x + i, y + j, 
                                   (row & 1) ? gc.foregroundColor : gc.backgroundColor);
      if (status > GLIB_ERROR_NOTHING_TO_DRAW) return (-1);
      row >>= 1;
    }
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str)
  \brief       Draw string (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   str    Null-terminated String
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str) {

  while (*str) {
    GLCD_DrawChar(x, y, *str++); x += pFont->width;
  }

  return 0;
}

/**
  \fn          int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val)
  \brief       Draw bargraph (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Full bargraph width in pixels
  \param[in]   height Full bargraph height in pixels
  \param[in]   val    Active bargraph value (maximum value is 100)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val) {
  GLIB_Rectangle_t rect;
  uint32_t b, c;
  EMSTATUS status;

  b = (val * width) / 100;

  rect.yMin = y;
  rect.yMax = y + height - 1;

  if ((width - b) > 0) {
    rect.xMin = x + b;
    rect.xMax = x + width  - 1;
    c = gc.foregroundColor;
    gc.foregroundColor = gc.backgroundColor;
    status = GLIB_drawRectFilled(&gc, &rect);
    gc.foregroundColor = c;
    if (status != GLIB_OK) {
      return -1;
    }
  }
  if (b > 0) {
    rect.xMin = x;
    rect.xMax = x + b - 1;
    status = GLIB_drawRectFilled(&gc, &rect);
    if (status != GLIB_OK) {
      return -1;
    }
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap)
  \brief       Draw bitmap (bitmap from BMP file without header)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Bitmap width in pixels
  \param[in]   height Bitmap height in pixels
  \param[in]   bitmap Bitmap data
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap) {
  return -1;
}

/**
  \fn          int32_t GLCD_VScroll (uint32_t dy)
  \brief       Scroll picture on display vertically
  \param[in]   dy  Scroll size in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_VScroll (uint32_t dy) {
  return -1;
}

/**
  \fn          int32_t GLCD_FrameBufferAccess (bool enable)
  \brief       Enable or disable direct access to FrameBuffer
  \param[in]   enable Enable or disable request
                - true (1): enable
                - false(0): disable
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t  GLCD_FrameBufferAccess (bool enable) {
#ifdef RTE_BOARD_SUPPORT_DRIVERS_TFT_RGB
  /* Direct data access is used with direct access to frame buffer */
  return 0;
#else
  return -1;
#endif
}

/**
  \fn          uint32_t GLCD_FrameBufferAddress (void)
  \brief       Retrieve FrameBuffer address
  \returns     FrameBuffer address
*/
uint32_t GLCD_FrameBufferAddress (void) {
  return ((uint32_t)frameBuffer);
}
