/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2017 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        19. September 2017
 * $Revision:    V2.4
 *
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file
 * Project:      USB Device Driver for SiliconLabs EFM32GGxxx
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                  Value
 *   ---------------------                  -----
 *   Connect to hardware via Driver_USBD# = 0
 * --------------------------------------------------------------------------
 * Defines used for driver configuration (at compile time):
 *
 *   USBD_MAX_ENDPOINT_NUM:  defines maximum number of IN/OUT Endpoint pairs 
 *                           that driver will support with Control Endpoint 0
 *                           not included, this value impacts driver memory
 *                           requirements
 *     - default value: 6
 *     - maximum value: 6
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 2.4
 *    Added support for CMSIS-RTOS2
 *  Version 2.3
 *    PowerControl for Power OFF and Uninitialize functions made unconditional
 *  Version 2.2
 *    Corrected IN ZLP sending
 *  Version 2.1
 *    Send data on first NAK to prevent simultaneous write to FIFO from library
 *    and from IRQ routine
 *  Version 2.0
 *    Updated to 2.00 API
 *  Version 1.4
 *    Multiple packet read
 *  Version 1.3
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.0
 *    Initial release
 */

/* NOTE:
 *  Isochronous Endpoints are not supported
 */

#include <stdint.h>
#include <string.h>

#include "RTE_Components.h"

#if       defined(RTE_CMSIS_RTOS2)
#include "cmsis_os2.h"
#elif     defined(RTE_CMSIS_RTOS)
#include "cmsis_os.h"
#endif

#include "Driver_USBD.h"

#include "em_device.h"
#include "em_cmu.h"

#ifndef USBD_MAX_ENDPOINT_NUM
#define USBD_MAX_ENDPOINT_NUM           6U
#endif
#if    (USBD_MAX_ENDPOINT_NUM > 6)
#error  Too many Endpoints, maximum IN/OUT Endpoint pairs that this driver supports is 6 !!!
#endif

#define USBD_DRIVER_INITIALIZED (1U     )
#define USBD_DRIVER_POWERED     (1U << 1)

extern uint8_t usb_role;
extern uint8_t usb_drv_state;

extern void USB_PinsConfigure   (uint8_t pins_mask);
extern void USB_PinsUnconfigure (uint8_t pins_mask);


// USBD Driver *****************************************************************

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(2,4)

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  0U,   // VBUS Detection
  0U,   // Event VBUS On
  0U    // Event VBUS Off
};

#define EP_NUM(ep_addr)         ((ep_addr) & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_ID(ep_addr)          ((EP_NUM(ep_addr) * 2U) + (((ep_addr) >> 7U) & 1U))

// FIFO sizes in bytes (total available memory for FIFOs is 2 kB)
#define USB_RX_FIFO_SIZE        896U
#define USB_TX0_FIFO_SIZE       192U
#define USB_TX1_FIFO_SIZE       192U
#define USB_TX2_FIFO_SIZE       192U
#define USB_TX3_FIFO_SIZE       192U
#define USB_TX4_FIFO_SIZE       192U
#define USB_TX5_FIFO_SIZE       192U
#define USB_TX6_FIFO_SIZE       192U

#define USB_TX_FIFO(n)          *((volatile uint32_t *)(USB->FIFO0D + (n * 1024U)))
#define USB_RX_FIFO             *((volatile uint32_t *)(USB->FIFO0D))

#define USB_DIEPTSIZ(ep_num)    *((volatile uint32_t *)(&USB->DIEP0TSIZ   + (ep_num * 8U)))
#define USB_DIEPCTL(ep_num)     *((volatile uint32_t *)(&USB->DIEP0CTL    + (ep_num * 8U)))
#define USB_DTXFSTS(ep_num)     *((volatile uint32_t *)(&USB->DIEP0TXFSTS + (ep_num * 8U)))
#define USB_DOEPTSIZ(ep_num)    *((volatile uint32_t *)(&USB->DOEP0TSIZ   + (ep_num * 8U)))
#define USB_DOEPCTL(ep_num)     *((volatile uint32_t *)(&USB->DOEP0CTL    + (ep_num * 8U)))
#define USB_DIEPINT(ep_num)     *((volatile uint32_t *)(&USB->DIEP0INT    + (ep_num * 8U)))
#define USB_DOEPINT(ep_num)     *((volatile uint32_t *)(&USB->DOEP0INT    + (ep_num * 8U)))

#define USB_EP_IN_TYPE(ep_num)  ((USB_DIEPCTL(ep_num) >> 18) & 3U)
#define USB_EP_OUT_TYPE(ep_num) ((USB_DOEPCTL(ep_num) >> 18) & 3U)

typedef struct {                        // Endpoint structure definition
  uint8_t  *data;
  uint32_t  num;
  uint32_t  num_transferred_total;
  uint16_t  num_transferring;
  uint16_t  max_packet_size;
  uint8_t   active;
  uint8_t   in_nak;
  uint8_t   in_zlp;
} ENDPOINT_t;

static ARM_USBD_SignalDeviceEvent_t   SignalDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t SignalEndpointEvent;

static ARM_USBD_STATE      usbd_state;
static uint32_t            setup_packet[2];     // Setup packet data
static volatile uint8_t    setup_received;      // Setup packet received

// Endpoints runtime information
static volatile ENDPOINT_t ep[(USBD_MAX_ENDPOINT_NUM + 1U) * 2U];

// Auxiliary functions

/**
  \fn          void USBD_FlushInEpFifo (uint8_t ep_addr)
  \brief       Flush IN Endpoint FIFO
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
*/
static void USBD_FlushInEpFifo (uint8_t ep_addr) {
  USB->GRSTCTL = (USB->GRSTCTL & ~_USB_GRSTCTL_TXFNUM_MASK)         |
                 (((uint32_t)ep_addr) << _USB_GRSTCTL_TXFNUM_SHIFT) |
                  USB_GRSTCTL_TXFFLSH;
}

/**
  \fn          void USBD_Reset (void)
  \brief       Reset USB Endpoint settings and variables.
*/
static void USBD_Reset (void) {
  uint8_t i;

  // Reset global variables
  setup_packet[0] = 0U;
  setup_packet[1] = 0U;
  setup_received  = 0U;
  memset((void *)&usbd_state, 0, sizeof(usbd_state));
  memset((void *)ep,          0, sizeof(ep));

  // Clear Endpoint mask registers
  USB->DOEPMSK = 0U;
  USB->DIEPMSK = 0U;

  for (i = 1U; i <= USBD_MAX_ENDPOINT_NUM; i++) {
    if ((USB_DOEPCTL(i) & USB_DOEP_CTL_EPENA) != 0U) {
      USB_DOEPCTL(i)   = USB_DOEP_CTL_EPDIS |       // Endpoint disable
                         USB_DOEP_CTL_SNAK;         // Endpoint set NAK
    }
    if ((USB_DIEPCTL(i) & USB_DIEP_CTL_EPENA) != 0U) {
      USB_DIEPCTL(i)   = USB_DIEP_CTL_EPDIS |       // Endpoint disable
                         USB_DIEP_CTL_SNAK;         // Endpoint set NAK
    }
    USBD_FlushInEpFifo(i);

    // Clear IN Endpoint interrupts
    USB_DIEPINT(i) = USB_DIEP_INT_XFERCOMPL   |
                     USB_DIEP_INT_EPDISBLD    |
                     USB_DIEP_INT_INTKNTXFEMP |
                     USB_DIEP_INT_INEPNAKEFF  |
                     USB_DIEP_INT_TXFEMP      ;

    // Clear OUT Endpoint interrupts
    USB_DOEPINT(i) = USB_DOEP_INT_XFERCOMPL   |
                     USB_DOEP_INT_EPDISBLD    |
                     USB_DOEP_INT_SETUP       |
                     USB_DOEP_INT_OUTTKNEPDIS ;
  }

  // Set device address to 0
  USB->DCFG       = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK);
  USB->DAINTMSK   =  USB_DAINTMSK_INEPMSK0    |     // Enable IN Endpoint0 interrupt
                     USB_DAINTMSK_OUTEPMSK0;        // Enable OUT Endpoint0 interrupt

  // Enable Setup phase done, OUT Endpoint disabled and OUT transfer complete interrupt
  USB->DOEPMSK    = USB_DOEPMSK_SETUPMSK      |
                    USB_DOEPMSK_XFERCOMPLMSK  ;

  // Enable In Endpoint disable and IN transfer complete interrupt
  USB->DIEPMSK    = USB_DIEPMSK_XFERCOMPLMSK;

  // Configure FIFOs
  USB->GRXFSIZ    =   USB_RX_FIFO_SIZE  / 4U;
  USB->GNPTXFSIZ  =  (USB_RX_FIFO_SIZE  / 4U) |
                    ((USB_TX0_FIFO_SIZE / 4U) << 16);

  USB->DIEPTXF1   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE) / 4U) |
                    ((USB_TX1_FIFO_SIZE / 4U) << 16);
  USB->DIEPTXF2   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE) / 4U) |
                    ((USB_TX2_FIFO_SIZE / 4U) << 16);
  USB->DIEPTXF3   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE +
                      USB_TX2_FIFO_SIZE)/ 4U) | ((USB_TX3_FIFO_SIZE / 4U) << 16);
  USB->DIEPTXF4   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE +
                      USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE) / 4U) | ((USB_TX4_FIFO_SIZE / 4U) << 16);
  USB->DIEPTXF5   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE +
                      USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE) / 4U) |
                    ((USB_TX5_FIFO_SIZE / 4U) << 16);
  USB->DIEPTXF6   = ((USB_RX_FIFO_SIZE  + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE +
                      USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE+
                      USB_TX5_FIFO_SIZE)/ 4U) | ((USB_TX6_FIFO_SIZE / 4U) << 16);
}

/**
  \fn          void USBD_EndpointReadSet (uint8_t ep_addr)
  \brief       Set Endpoint for next read.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
*/
static void USBD_EndpointReadSet (uint8_t ep_addr) {
  volatile ENDPOINT_t *ptr_ep;
  uint16_t             num;
  uint8_t              ep_num;

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  // Set packet count and transfer size
  if  (ptr_ep->num > ptr_ep->max_packet_size) { num = ptr_ep->max_packet_size; }
  else                                        { num = ptr_ep->num;             }

  if (ep_num != 0U) {
    USB_DOEPTSIZ(ep_num) = (1U << _USB_DOEP_TSIZ_PKTCNT_SHIFT) |
                            num                                ;
  } else {
    USB_DOEPTSIZ(0U)     = (1U << _USB_DOEP0TSIZ_PKTCNT_SHIFT) |
                           (3U << _USB_DOEP0TSIZ_SUPCNT_SHIFT) |
                            num                                ;
  }

  // Clear NAK and enable Endpoint
  USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_EPENA | USB_DOEP_CTL_CNAK;
}

/**
  \fn          int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint16_t num)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   num      number of data bytes to read
  \return      number of data bytes read
*/
static int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint16_t num) {
  volatile ENDPOINT_t *ptr_ep;
  uint32_t             i, residue;
  uint8_t             *ptr_dest_8;
  __packed uint32_t   *ptr_dest_32;
  volatile uint32_t   *ptr_src;
  uint8_t              ep_num;
  uint8_t              tmp_buf[4];

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  // Check if Endpoint is activated and buffer available
  if ((USB_DOEPCTL(ep_num) & USB_DOEP_CTL_USBACTEP) == 0U) { return 0U; }
  if (ptr_ep->data == 0U)                                  { return 0U; }

  if (num > ptr_ep->num) { num = ptr_ep->num; }

  // Copy data from FIFO
  ptr_src     = (volatile uint32_t *)(USB->FIFO0D);
  ptr_dest_32 = (__packed uint32_t *)(ptr_ep->data + ptr_ep->num_transferred_total);
  i           = num / 4U;
  while (i != 0U) {
    *ptr_dest_32++ = *ptr_src;
    i--;
  }
  ptr_ep->num_transferred_total += num;

  // If data size is not equal n*4
  residue = num % 4U;
  if (residue != 0U) {
    ptr_dest_8 = (uint8_t *)(ptr_dest_32);
    *((__packed uint32_t *)tmp_buf) = USB_RX_FIFO;
    for (i = 0U; i < residue; i++) {
      *ptr_dest_8++ = tmp_buf[i];
    }
  }

  if (num != ptr_ep->max_packet_size) { ptr_ep->num  = 0U;  }
  else                                { ptr_ep->num -= num; }

  return num;
}

/**
  \fn          void USBD_WriteToFifo (uint8_t ep_addr)
  \brief       Write data to Endpoint FIFO.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
*/
static void USBD_WriteToFifo (uint8_t ep_addr) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;
  uint16_t             num, i;
  volatile uint32_t   *ptr_dest;
  __packed uint32_t   *ptr_src;

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  if (ptr_ep->num > ptr_ep->max_packet_size) { num = ptr_ep->max_packet_size; }
  else                                       { num = ptr_ep->num;             }

  // Check if enough space in FIFO
  if ((USB_DTXFSTS(ep_num) * 4U) < num) { return; }

  // Set transfer size and packet count
  USB_DIEPTSIZ(ep_num) = (1U << _USB_DIEP_TSIZ_PKTCNT_SHIFT) |
                         (1U << _USB_DIEP_TSIZ_MC_SHIFT)     |
                          num                                ;

  // Enable Endpoint and clear NAK
  USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_EPENA | USB_DIEP_CTL_CNAK;

  ptr_src  = (__packed uint32_t *)(ptr_ep->data + ptr_ep->num_transferred_total);
  ptr_dest = (volatile uint32_t *)(USB->FIFO0D + (ep_num * 1024U));

  ptr_ep->num_transferring  = num;
  ptr_ep->num              -= num;
  ptr_ep->in_zlp            = 0U;

  // Copy data to FIFO
  i = (num + 3U) >> 2U;
  while (i != 0U) {
    *ptr_dest = *ptr_src++;
    i--;
  }
}


// USBD Driver functions

/**
  \fn          ARM_DRIVER_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                        ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  if ((usb_drv_state & USBD_DRIVER_INITIALIZED) != 0U) { return ARM_DRIVER_OK; }

  SignalDeviceEvent   =  cb_device_event;
  SignalEndpointEvent =  cb_endpoint_event;

  usb_role            =  ARM_USB_ROLE_DEVICE;
  USB_PinsConfigure  (ARM_USB_PIN_DP | ARM_USB_PIN_DM);

  usb_drv_state       =  USBD_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  USB_PinsUnconfigure (ARM_USB_PIN_DP | ARM_USB_PIN_DM);
  usb_role            =  ARM_USB_ROLE_NONE;
  usb_drv_state      &= ~USBD_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ      (USB_IRQn);                  // Disable interrupt
      NVIC_ClearPendingIRQ (USB_IRQn);                  // Clear pending interrupt
      usb_drv_state &= ~USBD_DRIVER_POWERED;            // Clear powered flag
      USB->GAHBCFG  &= ~USB_GAHBCFG_GLBLINTRMSK;        // Disable USB interrupts
                                                        // Reset variables
      setup_received =  0U;
      memset((void *)&usbd_state, 0, sizeof(usbd_state));
      memset((void *)ep,          0, sizeof(ep));

      USB->DCTL    |=  USB_DCTL_SFTDISCON;              // Soft disconnect enabled
      // Disable USB clock
      CMU->HFCORECLKEN0 &= ~(CMU_HFCORECLKEN0_USB | CMU_HFCORECLKEN0_USBC);
      break;

    case ARM_POWER_FULL:
      if ((usb_drv_state & USBD_DRIVER_POWERED) != 0U) { return ARM_DRIVER_OK; }

      CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USBC;
      CMU_ClockSelectSet (cmuClock_USBC, cmuSelect_HFCLK);
      CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USB;

      // Wait until AHB Master state machine is in the idle condition
      while ((USB->GRSTCTL & USB_GRSTCTL_AHBIDLE) == 0U);

      USB->GRSTCTL  |=  USB_GRSTCTL_CSFTRST;            // Core soft reset
      while ((USB->GRSTCTL & USB_GRSTCTL_CSFTRST) != 0U);
      osDelay(1U);

      // Wait until AHB Master state machine is in the idle condition
      while ((USB->GRSTCTL & USB_GRSTCTL_AHBIDLE) == 0U);

      USBD_Reset ();                                    // Reset variables and endpoint settings

      USB->DCTL     |=  USB_DCTL_SFTDISCON;             // Soft disconnect enabled

      // Set turnaround time
      USB->GUSBCFG   = ((USB->GUSBCFG & ~_USB_GUSBCFG_USBTRDTIM_MASK) |
                        (5U << _USB_GUSBCFG_USBTRDTIM_SHIFT))         ;
      if (((USB->GUSBCFG & USB_GUSBCFG_FORCEDEVMODE) == 0U) || ((USB->GUSBCFG & USB_GUSBCFG_FORCEHSTMODE) == 1U)) {
        USB->GUSBCFG &= ~USB_GUSBCFG_FORCEHSTMODE;
        USB->GUSBCFG |=  USB_GUSBCFG_FORCEDEVMODE;
        osDelay(50U);
      }

      USB->GINTMSK   = (USB_GINTSTS_USBSUSP      |      // Unmask interrupts
                        USB_GINTSTS_USBRST       |
                        USB_GINTSTS_ENUMDONE     |
                        USB_GINTSTS_RXFLVL       |
                        USB_GINTSTS_IEPINT       |
                        USB_GINTSTS_OEPINT       |
                        USB_GINTSTS_WKUPINT)     ;

      USB->GAHBCFG  |= (USB_GAHBCFG_GLBLINTRMSK  |
                        USB_GAHBCFG_NPTXFEMPLVL) ;

      usb_drv_state |=  USBD_DRIVER_POWERED;            // Set powered flag
      NVIC_EnableIRQ       (USB_IRQn);                  // Enable interrupt
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB->DCTL  &= ~USB_DCTL_SFTDISCON;    // Soft disconnect disabled

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB->DCTL  |=  USB_DCTL_SFTDISCON;    // Soft disconnect enabled

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB->DCTL |=   USB_DCTL_RMTWKUPSIG;   // Remote wakeup signaling
  osDelay(5U);
  USB->DCTL &=  ~USB_DCTL_RMTWKUPSIG;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB->DCFG = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | (dev_addr << _USB_DCFG_DEVADDR_SHIFT) ;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }
  if (setup_received                        == 0U) { return ARM_DRIVER_ERROR; }

  setup_received = 0U;
  memcpy(setup, setup_packet, 8);

  if (setup_received != 0U) {           // If new setup packet was received while this was being read
    return ARM_DRIVER_ERROR;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                               uint8_t  ep_type,
                                               uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;
  uint16_t             ep_mps;
  bool                 ep_dir;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)              { return ARM_DRIVER_ERROR; }
  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ptr_ep = &ep[EP_ID(ep_addr)];
  if (ptr_ep->active != 0U)                        { return ARM_DRIVER_ERROR_BUSY; }

  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;
  ep_mps =  ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

  // Clear Endpoint transfer and configuration information
  memset((void *)(ptr_ep), 0, sizeof (ENDPOINT_t));

  // Set maximum packet size to requested
  ptr_ep->max_packet_size = ep_mps;

  if (ep_dir != 0U) {                                   // IN Endpoint
    ptr_ep->in_nak = 0U;                                // Clear IN Endpoint NAK flag

    // Configure IN Endpoint
    USB_DIEPCTL(ep_num) = (ep_num  << _USB_DIEP_CTL_TXFNUM_SHIFT) |     // FIFO Number
                          (ep_type << _USB_DIEP_CTL_EPTYPE_SHIFT) |     // Endpoint Type
                           ep_mps                                 ;     // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_SETD0PIDEF;
    }

    USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_USBACTEP;     // Activate Endpoint

    if ((USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA) != 0U) {
      USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_EPDIS;        // Disable Endpoint
    }

    USBD_FlushInEpFifo (ep_addr);
  } else {                              // OUT Endpoint

    // Configure OUT Endpoint
    USB_DOEPCTL(ep_num) = (ep_type <<  _USB_DOEP_CTL_EPTYPE_SHIFT) |    // Endpoint Type
                           USB_DOEP_CTL_SNAK                       |    // Set NAK
                           ep_mps                                  ;    // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SETD0PIDEF;
    }

    USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_USBACTEP;       // Activate Endpoint
  }

  USB->DAINTMSK |= (1U << (ep_num + ((ep_dir ^ 1U) << 4U)));    // Enable Endpoint interrupt

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;
  bool                 ep_dir;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)              { return ARM_DRIVER_ERROR; }
  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ptr_ep = &ep[EP_ID(ep_addr)];
  if (ptr_ep->active != 0U)                         { return ARM_DRIVER_ERROR_BUSY; }

  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  USB->DAINTMSK &= ~(1U << (ep_num + ((ep_dir ^ 1U) << 4U)));           // Disable Endpoint interrupt

  // Clear Endpoint transfer and configuration information
  memset((void *)(ptr_ep), 0, sizeof (ENDPOINT_t));

  if (ep_dir != 0U) {                                   // IN Endpoint

    if ((USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA) != 0U) {
      USB_DIEPCTL(ep_num)  |=  USB_DIEP_CTL_EPDIS;      // Disable Endpoint
    }

    USB_DIEPCTL(ep_num)    |=  USB_DIEP_CTL_SNAK;       // Set Endpoint NAK

    if (ep_num != 0U) { USB_DIEPCTL(ep_num) &= ~USB_DIEP_CTL_USBACTEP; }        // Deactivate Endpoint
  } else {                              // OUT Endpoint
    USB->DCTL |= USB_DCTL_SGOUTNAK;                     // Set Global OUT NAK
    while ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0U);

    USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SNAK;           // Set Endpoint NAK

    if (ep_num != 0U) {
      // Disable OUT Endpoint
      if ((USB_DOEPCTL(ep_num) & USB_DOEP_CTL_EPENA) != 0U) {   // If Endpoint is Enabled
        USB_DOEPCTL(ep_num)  |=  USB_DOEP_CTL_EPDIS;    // Disable Endpoint
        while (!(USB_DOEPINT(ep_num) & USB_DOEP_INT_EPDISBLD));
      }
      USB_DOEPCTL(ep_num) &= ~USB_DOEP_CTL_USBACTEP;    // Deactivate Endpoint
    }

    USB->DCTL |= USB_DCTL_CGOUTNAK;                     // Clear Global OUT NAK
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;
  bool                 ep_dir;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)              { return ARM_DRIVER_ERROR; }
  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ptr_ep = &ep[EP_ID(ep_addr)];
  if (ptr_ep->active != 0U)                        { return ARM_DRIVER_ERROR_BUSY; }

  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  if (stall != 0U) {                                    // Activate STALL
    if (ep_dir != 0U) {                                 // IN Endpoint
      if ((USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA) != 0U) {
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL | USB_DIEP_CTL_EPDIS;
      } else {
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL;
      }

      USBD_FlushInEpFifo (ep_addr);
    } else {                                            // OUT Endpoint
      USB->DCTL |= USB_DCTL_SGOUTNAK;                   // Set Global OUT NAK
      while ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0U);

      // Stall OUT Endpoint
      if ((USB_DOEPCTL(ep_num) & USB_DOEP_CTL_EPENA) != 0U) {
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL | USB_DOEP_CTL_EPDIS;
      } else {
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL;
      }

      USB->DCTL |= USB_DCTL_CGOUTNAK;                   // Clear global NAK
    }
  } else {                                              // Clear STALL
    ptr_ep->in_nak = 0U;
    ptr_ep->in_zlp = 0U;
    if (ep_dir != 0U) {                                 // IN Endpoint
      if ((USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA) != 0U) {  // If Endpoint enabled
        USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS;    // Disable Endpoint
      }

      USBD_FlushInEpFifo (ep_addr);

      // Set DATA0 PID for Interrupt and Bulk Endpoint
      if (((USB_DIEPCTL(ep_num) & _USB_DIEP_CTL_EPTYPE_MASK) >> _USB_DIEP_CTL_EPTYPE_SHIFT) > 1) {
        USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_SETD0PIDEF;
      }

      USB_DIEPCTL(ep_num) &= ~USB_DIEP_CTL_STALL;       // Clear Stall
    } else {                                            // Clear OUT Endpoint stall
      // Set DATA0 pid for Interrupt and Bulk Endpoint
      if (((USB_DOEPCTL(ep_num) & _USB_DOEP_CTL_EPTYPE_MASK) >> _USB_DOEP_CTL_EPTYPE_SHIFT) > 1) {
        USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SETD0PIDEF;
      }
      USB_DOEPCTL(ep_num) &= ~USB_DOEP_CTL_STALL;       // Clear Stall
    }
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;
  bool                 ep_dir;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)              { return ARM_DRIVER_ERROR; }
  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ptr_ep = &ep[EP_ID(ep_addr)];
  if (ptr_ep->active != 0U)                        { return ARM_DRIVER_ERROR_BUSY; }

  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  ptr_ep->active = 1U;

  ptr_ep->data                  = data;
  ptr_ep->num                   = num;
  ptr_ep->num_transferred_total = 0U;
  ptr_ep->num_transferring      = 0U;

  if (ep_dir != 0U) {                                   // IN Endpoint
    if (num == 0U) {
      ptr_ep->in_zlp     =  1U;                         // Send IN ZLP requested
    }
    ptr_ep->in_nak       =  1U;                         // Set IN Endpoint NAK flag
    USB_DIEPCTL(ep_num) |=  USB_DIEP_CTL_CNAK;          // Clear NAK
    USB_DIEPCTL(ep_num) |=  USB_DIEP_CTL_SNAK;          // Set NAK
    USB->DIEPMSK        |=  USB_DIEPMSK_INEPNAKEFFMSK;  // Enable NAK effective interrupt
  } else {                                              // OUT Endpoint
    USBD_EndpointReadSet(ep_addr);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transferred data bytes
*/
static uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {

  if (EP_NUM(ep_addr) > USBD_MAX_ENDPOINT_NUM) { return 0U; }

  return (ep[EP_ID(ep_addr)].num_transferred_total);
}

/**
  \fn          int32_t USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  volatile ENDPOINT_t *ptr_ep;
  uint8_t              ep_num;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)              { return ARM_DRIVER_ERROR; }
  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ptr_ep = &ep[EP_ID(ep_addr)];

  ptr_ep->num    = 0U;
  ptr_ep->in_nak = 0U;
  ptr_ep->in_zlp = 0U;

  if ((ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) != 0U) {
    if ((USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA) != 0U) {    // If Endpoint enabled
      USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS;      // Disable Endpoint
    }

    USB_DIEPCTL(ep_num)     |= USB_DIEP_CTL_SNAK;       // Set NAK

    USBD_FlushInEpFifo (ep_addr);
  } else {
    if (USB_DOEPCTL(ep_num) &  USB_DOEP_CTL_EPENA) {    // If Endpoint enabled
      USB_DOEPCTL(ep_num)   |= USB_DOEP_CTL_EPDIS;      // Disable Endpoint
    }

    USB_DOEPCTL(ep_num)     |= USB_DOEP_CTL_SNAK;       // Set NAK
  }

  ptr_ep->active = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {

  if ((usb_drv_state & USBD_DRIVER_POWERED) == 0U) { return 0U; }

  return ((USB->DSTS & _USB_DSTS_SOFFN_MASK) >> _USB_DSTS_SOFFN_SHIFT);
}

/**
  \fn          void USBD_IRQ (uint32_t gintsts)
  \brief       USB Device Interrupt Routine (IRQ).
*/
void USBD_IRQ (uint32_t gintsts) {
  volatile ENDPOINT_t *ptr_ep, *ptr_ep_in;
  uint32_t             val, msk, ep_int;
  uint16_t             num;
  uint8_t              ep_num, i;

  if ((gintsts & USB_GINTSTS_USBRST) != 0U) {           // Reset interrupt
    USBD_Reset();
    SignalDeviceEvent(ARM_USBD_EVENT_RESET);
    USB->GINTSTS =  USB_GINTSTS_USBRST;
  }

  if ((gintsts & USB_GINTSTS_USBSUSP) != 0U) {          // Suspend interrupt
    usbd_state.active = 0U;
    SignalDeviceEvent(ARM_USBD_EVENT_SUSPEND);
    USB->PCGCCTL |=  USB_PCGCCTL_STOPPCLK;              // Stop PHY clock
    USB->GINTSTS  =  USB_GINTSTS_USBSUSP;
  }

  if ((gintsts & USB_GINTSTS_WKUPINT) != 0U) {          // Resume interrupt
    usbd_state.active = 1U;
    SignalDeviceEvent(ARM_USBD_EVENT_RESUME);
    USB->PCGCCTL &= ~USB_PCGCCTL_STOPPCLK;              // Start PHY clock
    USB->GINTSTS  =  USB_GINTSTS_WKUPINT;
  }

  if ((gintsts & USB_GINTSTS_ENUMDONE) != 0U) {         // Speed enumeration completed
    usbd_state.speed  = ARM_USB_SPEED_FULL;
    usbd_state.active = 1U;
    USB->DCTL    |= USB_DCTL_CGNPINNAK;                 // Clear global IN NAK
    USB->DCTL    |= USB_DCTL_CGOUTNAK;                  // Clear global OUT NAK
    USB->GINTSTS  = USB_GINTSTS_ENUMDONE;
  }

  if ((gintsts & USB_GINTSTS_RXFLVL) != 0U) {           // Receive FIFO interrupt
    val    =  USB->GRXSTSP;
    ep_num =  val & 0x0FU;
    num    = (val >> 4U) & 0x7FFU;
    switch ((val >> 17U) & 0x0FU) {
      case 6:                                           // Setup packet
        // Read setup packet
        setup_packet[0] = USB_RX_FIFO;
        setup_packet[1] = USB_RX_FIFO;

        // Analyze Setup packet for SetAddress
        if ((setup_packet[0] & 0xFFFFU) == 0x0500U) {
          USBD_DeviceSetAddress((setup_packet[0] >> 16) & 0xFFU);
        }
        setup_received = 1U;
        break;
      case 2:                                           // OUT packet
        USBD_ReadFromFifo(ep_num, num);
        break;
      case 1:                                           // Global OUT NAK
      case 3:                                           // OUT transfer completed
      case 4:                                           // SETUP transaction completed
      default:
        break;
    }
  }

  if ((gintsts & USB_GINTSTS_OEPINT) != 0U) {           // If OUT Packet
    msk    = (((USB->DAINT & USB->DAINTMSK) >> 16) & 0xFFFFU);
    ep_num = 0U;
    do {
      if (((msk >> ep_num) & 1U) != 0U) {
        ep_int = USB_DOEPINT(ep_num) & USB->DOEPMSK;
        ptr_ep = &ep[EP_ID(ep_num)];
        if ((ep_int & USB_DOEP_INT_SETUP) != 0U) {      // If setup phase done interrupt
          ptr_ep->num = 0U;
          USB_DOEPINT(ep_num) = USB_DOEP_INT_SETUP;
          SignalEndpointEvent(ep_num, ARM_USBD_EVENT_SETUP);
        }

        if ((ep_int & USB_DOEP_INT_XFERCOMPL) != 0U) {  // If transfer complete interrupt
          USB_DOEPINT(ep_num) = USB_DOEP_INT_XFERCOMPL;
          if (USB_EP_OUT_TYPE(ep_num) != ARM_USB_ENDPOINT_ISOCHRONOUS) {
            if (ptr_ep->num != 0U) {
              USBD_EndpointReadSet(ep_num);
            } else {
              ptr_ep->active = 0U;
              SignalEndpointEvent(ep_num, ARM_USBD_EVENT_OUT);
            }
          }
        }
      }
      ep_num++;
    } while ((msk >> ep_num) != 0U);
  }

  if ((gintsts & USB_GINTSTS_IEPINT) != 0U) {           // If IN Packet
    msk    = (USB->DAINT & USB->DAINTMSK & 0xFFFFU);
    ep_num = 0U;
    do {
      if (((msk >> ep_num) & 1U) != 0U) {
        ep_int = USB_DIEPINT(ep_num) & USB->DIEPMSK;
        ptr_ep = &ep[EP_ID(ep_num | ARM_USB_ENDPOINT_DIRECTION_MASK)];
        if ((ep_int & USB_DIEP_INT_INEPNAKEFF) != 0U) { // If IN Endpoint NAK effective
          if (ptr_ep->in_nak != 0U) {
            ptr_ep->in_nak = 0U;
            val = 0U;
            ptr_ep_in = &ep[0];
            for (i = 0U; i < USBD_MAX_ENDPOINT_NUM; i++) {
              val |= ptr_ep_in->in_nak;
              ptr_ep_in += 2U;
            }

            // If no more forced NAKs, disable IN NAK effective interrupt
            if (val == 0U) { USB->DIEPMSK &= ~USB_DIEPMSK_INEPNAKEFFMSK; }

            // If Data available, write Data
            if ((ptr_ep->num != 0U) || (ptr_ep->in_zlp != 0U)) {
              USBD_WriteToFifo(ep_num | ARM_USB_ENDPOINT_DIRECTION_MASK);
            }
          }
          USB_DIEPINT(ep_num) = USB_DIEP_INT_INEPNAKEFF;
        }

        if ((ep_int & USB_DIEP_INT_XFERCOMPL) != 0U) {  // If transmit completed
          USB_DIEPINT(ep_num) = USB_DIEP_INT_XFERCOMPL;
          ptr_ep->num_transferred_total += ptr_ep->num_transferring;
          if (ptr_ep->num == 0U) {
            ptr_ep->data   = NULL;
            ptr_ep->active = 0U;
            SignalEndpointEvent(ep_num | ARM_USB_ENDPOINT_DIRECTION_MASK, ARM_USBD_EVENT_IN);
          } else {
            USBD_WriteToFifo(ep_num | ARM_USB_ENDPOINT_DIRECTION_MASK);
          }
        }
      }
      ep_num++;
    } while ((msk >> ep_num != 0U));
  }
}

ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};
