/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2016 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        6. September 2016
 * $Revision:    V1.4
 *
 * Project:      USB common (Device and Host) module for SiliconLabs EFM32GGxxx
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.4
 *    Corrected Host overcurrent pin functionality
 *  Version 1.3
 *    Corrected IRQ handler routine for Host and Device capable driver
 *  Version 1.2
 *    Removed configured pins checking
 *  Version 1.1
 *    Removed __weak function definitions
 *  Version 1.0
 *    Initial release
 */

#include "em_device.h"

#include "em_cmu.h"
#include "em_emu.h"
#include "em_gpio.h"

#include "Driver_USB.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

#if (RTE_USB)
#if (RTE_USB_HOST && !RTE_USB_VBUS_PIN)
#error   Configure VBUS Power On/Off pin for USB OTG Full-speed in RTE_Device.h!
#endif
#else
#error   Enable USB Full-speed in RTE_Device.h!
#endif

extern void USBH_IRQ (uint32_t gintsts);
extern void USBD_IRQ (uint32_t gintsts);

uint8_t usb_role      = ARM_USB_ROLE_NONE;
uint8_t usb_drv_state = 0U;


// Common IRQ Routine **********************************************************

/**
  \fn          void USB_IRQHandler (void)
  \brief       USB Interrupt Routine (IRQ).
*/
void USB_IRQHandler (void) {
  uint32_t gintsts;

  gintsts = USB->GINTSTS & USB->GINTMSK;

#if (defined(RTE_Drivers_USBH0) && defined(RTE_Drivers_USBD0))
  switch (usb_role) {
#ifdef RTE_Drivers_USBH0
    case ARM_USB_ROLE_HOST:
      USBH_IRQ (gintsts);
      break;
#endif
#ifdef RTE_Drivers_USBD0
    case ARM_USB_ROLE_DEVICE:
      USBD_IRQ (gintsts);
      break;
#endif
    default:
      break;
  }
#else
#ifdef RTE_Drivers_USBH0
  USBH_IRQ (gintsts);
#else
  USBD_IRQ (gintsts);
#endif
#endif
}


// Public Functions ************************************************************

/**
  \fn          void USB_PinsConfigure (uint8_t pins_mask)
  \brief       Configure single or multiple USB Pin(s).
  \param[in]   Mask of pins to be configured (possible masking values:
               ARM_USB_PIN_DP, ARM_USB_PIN_DM, ARM_USB_PIN_VBUS,
               ARM_USB_PIN_OC, ARM_USB_PIN_ID)
*/
void USB_PinsConfigure (uint8_t pins_mask) {

  if ((pins_mask & (ARM_USB_PIN_DP | ARM_USB_PIN_DM)) != 0U) {
    // Init PHY. Enable USB pins (DP and DM)
    CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE        |= USB_ROUTE_PHYPEN;
  }
#if (defined(RTE_Drivers_USBH0) && defined(RTE_USB_HOST) && (RTE_USB_HOST == 1))
  if (usb_role == ARM_USB_ROLE_HOST) {
#if (defined(RTE_USB_VBUS_PIN) && (RTE_USB_VBUS_PIN == 1))
    if ((pins_mask & ARM_USB_PIN_VBUS) != 0U) {
      // Enable Host VBUS pin
      CMU_ClockEnable ( cmuClock_GPIO, true );
      GPIO_PinModeSet ( GPIO_PORT(RTE_USB_VBUS_PORT), RTE_USB_VBUS_BIT, gpioModePushPull, 0);
      CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USB;
      USB->ROUTE        |= USB_ROUTE_VBUSENPEN;
    }
#endif
#if (defined(RTE_USB_OC_PIN) && (RTE_USB_OC_PIN == 1))
    if ((pins_mask & ARM_USB_PIN_OC) != 0U) {
      // Enable overcurrent pin
      GPIO_PinModeSet (GPIO_PORT(RTE_USB_OC_PORT), RTE_USB_OC_BIT, gpioModeInput, 0);
    }
#endif
  }
#endif
}

/**
  \fn          void USB_PinsUnconfigure (uint8_t pins_mask)
  \brief       De-configure to reset settings single or multiple USB Pin(s).
  \param[in]   Mask of pins to be de-configured (possible masking values:
               ARM_USB_PIN_DP, ARM_USB_PIN_DM, ARM_USB_PIN_VBUS,
               ARM_USB_PIN_OC, ARM_USB_PIN_ID)
*/
void USB_PinsUnconfigure (uint8_t pins_mask) {

  if ((pins_mask & (ARM_USB_PIN_DP | ARM_USB_PIN_DM)) != 0U) {
    // Disable USB pins (DP and DM)
    USB->ROUTE    &=  ~USB_ROUTE_PHYPEN;
  }
#if (defined(RTE_Drivers_USBH0) && defined(RTE_USB_HOST) && (RTE_USB_HOST == 1))
  if (usb_role == ARM_USB_ROLE_HOST) {
#if (defined(RTE_USB_VBUS_PIN) && (RTE_USB_VBUS_PIN == 1))
    if ((pins_mask & ARM_USB_PIN_VBUS) != 0U) {
      // Disable Host VBUS pin
      USB->ROUTE    &= ~USB_ROUTE_VBUSENPEN;
      GPIO_PinModeSet (GPIO_PORT(RTE_USB_VBUS_PORT), RTE_USB_VBUS_BIT, gpioModeDisabled, 0);
    }
#endif
#if (defined(RTE_USB_OC_PIN) && (RTE_USB_OC_PIN == 1))
    if ((pins_mask & ARM_USB_PIN_OC) != 0U) {
      // Disable overcurrent pin
      GPIO_PinModeSet (GPIO_PORT(RTE_USB_OC_PORT), RTE_USB_OC_BIT, gpioModeDisabled, 0);
    }
#endif
  }
#endif
}

/**
  \fn          void USB_PinVbusOnOff (bool state)
  \brief       Drive VBUS Pin On/Off.
  \param[in]   state    State On/Off (true = On, false = Off)
*/
void USB_PinVbusOnOff (bool state) {
}

/**
  \fn          bool USB_PinGetOC (void)
  \brief       Get state of Overcurrent Pin.
  \return      overcurrent state (true = Overcurrent active, false = No overcurrent)
*/
bool USB_PinGetOC (void) {
#if (defined(RTE_Drivers_USBH0) && defined(RTE_USB_HOST) && (RTE_USB_HOST == 1) && defined(RTE_USB_OC_PIN) && (RTE_USB_OC_PIN == 1))
#if (RTE_USB_OC_ACTIVE == 0)
  return ((GPIO_PortInGet(GPIO_PORT(RTE_USB_OC_PORT)) & (1U << RTE_USB_OC_BIT)) ? false : true);
#else
  return ((GPIO_PortInGet(GPIO_PORT(RTE_USB_OC_PORT)) & (1U << RTE_USB_OC_BIT)) ? true : false);
#endif
#else
  return false;
#endif
}
